#include "typedef.h"

#include <stdio.h>
#include <string.h>

#include "object.h"
#include "util.h"
#include "variable.h"
#include "class.h"
#include "edge.h"

Cvariable	*Cvariable::g_hash[HASHSIZE] = {0};

extern int g_zero_indirections;

/* static */ u4T
Cvariable::hash(Centity *entityP, const u2T *nameP)
{
	int		ret  = (int) entityP;
	u2T		c;
	const	u2T *P;

	for (P = nameP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	return(ret & (HASHSIZE-1));
}

Cvariable::Cvariable(Centity *parentP, const u2T *nameP, u2T *descriptorP)
{
	Cclass	*classP;
	u2T		*P;

	m_hashP          = 0;
	m_parentP        = parentP;
	m_nameP          = unicodedup(nameP);
	m_descriptorP    = unicodedup(descriptorP);
	m_access_flags   = 0;
	add_created();

	for (P = descriptorP; *P == '['; ++P);
	if (*P == 'L') {
		// Only create edges to classes or arrays of classes
		if (!g_zero_indirections) {
			P = descriptorP;
		}
		classP = Cclass::locate(P);
		new Cedge(this, classP, INSTANTIATES_EDGE, -1);
	}
}

Cvariable *
Cvariable::locate(Centity *parentP, const u2T *nameP, u2T *descriptorP)
{
	unsigned int	index;
	Cvariable		**variablePP, *variableP;

	index         = hash(parentP, nameP);
	for (variablePP = g_hash + index; variableP = *variablePP; variablePP = &(variableP->m_hashP)) {
		if (variableP->m_parentP != parentP) {
			continue;
		}
		if (unicodecmp(variableP->m_nameP, nameP)) {
			continue;
		}
		return(variableP);
	}	
	variableP = new Cvariable(parentP, nameP, descriptorP);
	if (!variableP) {
		outofmemory();
	}
	variableP->add_created();
	*variablePP = variableP;
	return(variableP);
}

/* virtual */ objectE
Cvariable::type(void) const
{
	return(variableE);
}

/* virtual */ void
Cvariable::emit_id(void) const
{
	fprintf(stdout, "L%p", this);
}

const char *
Cvariable::get_type(void) const
{
	return("L");
}
		
/* virtual */ void
Cvariable::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE L%p L\n", this);
}

void
Cvariable::emit_attributes(void) const
{
	emit_id();
	fputs(" {label=\"", stdout);
	dumpUnicode(m_nameP);
	fputc('"', stdout);

	fputs(" file=.", stdout);

	if (m_descriptorP) {
		fputs(" descriptor=\"", stdout);
		dump_descriptor(m_descriptorP);
		fputc('"', stdout);
	}

	dump_access(m_access_flags);
	fputs("}\n", stdout);
}

